/****************************************************************************
** Dooble - The Secure Internet Web Browser
**
** Copyright (c) 2008, 2009, 2010, 2011, 2012 Alexis Megas,
** Gunther van Dooble, and the Dooble Team.
** All rights reserved.
**
** License: GPL2 only:
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; version 2 of the License only.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
** or see here: http://www.gnu.org/licenses/gpl.html
**
** For the WebKit library, please see: http://webkit.org.
**
** THE CODE IS PROVIDED BY THE AUTHORS ''AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
** IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
** PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
** ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
** DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
** GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
** INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
** IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
** ARISING IN ANY WAY OUT OF THE USE OF THIS APPLICATION, EVEN IF ADVISED
** OF THE POSSIBILITY OF SUCH DAMAGE.
**
** Please report all praise, requests, bugs, and problems to the project
** team and administrators: http://sf.net/projects/dooble.
**
** You can find us listed at our project page. New team members are welcome.
** The name of the authors should not be used to endorse or promote products
** derived from Dooble without specific prior written permission.
** If you use this code for other projects, please let us know.
**
** Web sites:
**   http://sf.net/projects/dooble
**   http://dooble.sf.net
****************************************************************************/

#include <QDir>
#include <QtCore>
#include <QTimer>
#include <QSqlQuery>
#include <QProgressBar>
#include <QSqlDatabase>

#include "dmisc.h"
#include "dooble.h"
#include "dcookies.h"

dcookies::dcookies(void):QNetworkCookieJar()
{
  /*
  ** settingsWindow/cookiesShouldBe
  ** 0 - deleted upon exit
  ** 1 - kept forever (that's a long time)
  ** 2 - kept until they expire
  */

  m_timer = new QTimer(this);
  m_deleteAndSaveCookiesTimer = 0;
  connect(m_timer,
	  SIGNAL(timeout(void)),
	  this,
	  SLOT(slotRemoveDomains(void)));
  connect(this,
	  SIGNAL(cookieReceived(const QString &,
				const QUrl &,
				const QDateTime &)),
	  dooble::s_cookiesBlockWindow,
	  SLOT(slotAdd(const QString &,
		       const QUrl &,
		       const QDateTime &)));

  int interval = computedTimerInterval();

  if(interval > 0)
    m_timer->start(interval);

  createCookieDatabase();
}

dcookies::~dcookies()
{
  if(dooble::s_settings.value("settingsWindow/cookiesShouldBe",
			      0).toInt() == 2)
    deleteExpiredCookies();
  else
    purge();
}

QList<QNetworkCookie> dcookies::allCookies(void) const
{
  /*
  ** If cookies are disabled, should this method return an empty
  ** list or the current list? For now, it will return the
  ** current list of cookies minus the expired cookies.
  */

  QList<QNetworkCookie> list(QNetworkCookieJar::allCookies());

  if(dooble::s_settings.value("settingsWindow/cookiesShouldBe",
			      0).toInt() == 2)
    {
      QDateTime now(QDateTime::currentDateTime());

      /*
      ** Remove expired cookies.
      */

      for(int i = list.size() - 1; i >= 0; i--)
	if(list.at(i).isSessionCookie())
	  continue;
	else if(list.at(i).expirationDate().toLocalTime() <= now)
	  list.removeAt(i);
    }

  return list;
}

QList<QNetworkCookie> dcookies::cookiesForUrl(const QUrl &url) const
{
  /*
  ** If cookies are disabled, should this method return an empty
  ** list or the current list? For now, it will return the
  ** current list of cookies minus the expired cookies.
  */

  QList<QNetworkCookie> list(QNetworkCookieJar::cookiesForUrl(url));

  if(dooble::s_settings.value("settingsWindow/cookiesShouldBe",
			      0).toInt() == 2)
    {
      QDateTime now(QDateTime::currentDateTime());

      /*
      ** Remove expired cookies.
      */

      for(int i = list.size() - 1; i >= 0; i--)
	if(list.at(i).isSessionCookie())
	  continue;
	else if(list.at(i).expirationDate().toLocalTime() <= now)
	  list.removeAt(i);
    }

  return list;
}

bool dcookies::setCookiesFromUrl(const QList<QNetworkCookie> &cookieList,
				 const QUrl &url)
{
  if(!dooble::s_settings.value("settingsWindow/cookiesEnabled",
			       true).toBool())
    {
      if(m_deleteAndSaveCookiesTimer)
	m_deleteAndSaveCookiesTimer->stop();

      return false;
    }
  else if(cookieList.isEmpty())
    return false;

  /*
  ** Other exceptions models are only populated if some
  ** action has been denied. With cookies, knowledge of the
  ** cookie may be beneficial.
  */

  QDateTime now(QDateTime::currentDateTime());

  emit cookieReceived(url.host(), url, now);

  if(dooble::s_cookiesBlockWindow->allowed(url.host()))
    {
      emit exceptionRaised(dooble::s_cookiesBlockWindow, url);
      return false;
    }

  m_deleteCookies.append(QNetworkCookieJar::cookiesForUrl(url));

  /*
  ** Now proceed with adding new cookies to the jar.
  */

  bool added = false;
  bool acceptThirdParty = dooble::s_settings.value
    ("settingsWindow/acceptThirdPartyCookies", true).toBool();

  for(int i = 0; i < cookieList.size(); i++)
    {
      /*
      ** cookieList.at(i)'s domain may not be the same domain
      ** as url's.
      */

      if(!acceptThirdParty)
	if(url.host() != cookieList.at(i).domain())
	  if(cookieList.at(i).domain() != "" &&
	     cookieList.at(i).domain() != ".")
	    /*
	    ** Sorry, but third-party cookies are not
	    ** allowed. Goodbye!
	    */

	    continue;

      QList<QNetworkCookie> cookie;

      cookie.append(cookieList.at(i));

      if(!QNetworkCookieJar::setCookiesFromUrl(cookie, url))
	{
	  QList<QNetworkCookie> all(allCookies());

	  for(int j = all.size() - 1; j >= 0; j--)
	    if(cookie.at(0).domain() == all.at(j).domain() &&
	       cookie.at(0).name() == all.at(j).domain() &&
	       cookie.at(0).path() == all.at(j).path())
	      {
		/*
		** cookie.at(0) already exists. Let's remove it before
		** adding it.
		*/

		deleteCookie(all.at(j));
		all.removeAt(j);
		break;
	      }

	  all.append(cookie.at(0));
	  setAllCookies(all);
	  added = true;
	}
      else
	added = true;
    }

  m_saveCookies.append(cookiesForUrl(url));

  if(m_deleteAndSaveCookiesTimer)
    if(m_deleteCookies.size() > 0 || m_saveCookies.size() > 0)
      if(!m_deleteAndSaveCookiesTimer->isActive())
	m_deleteAndSaveCookiesTimer->start();

  if(added)
    emit changed();

  return added;
}

void dcookies::removeDomain(const QString &domain)
{
  QList<QNetworkCookie> all(allCookies());

  for(int i = all.size() - 1; i >= 0; i--)
    if(all.at(i).domain() == domain)
      /*
      ** Remove all cookies from the specified domain.
      */

      all.removeAt(i);

  setAllCookies(all);
  all.clear();
  m_favorites.remove(domain);

  if(dmisc::passphraseWasAuthenticated())
    {
      {
	QSqlDatabase db = QSqlDatabase::addDatabase("QSQLITE", "cookies");

	db.setDatabaseName(dooble::s_homePath +
			   QDir::separator() + "cookies.db");

	if(db.open())
	  {
	    QString hashType
	      (dooble::s_settings.value("settingsWindow/hashType").
	       toString());
	    QSqlQuery query(db);

	    query.exec("PRAGMA synchronous = OFF");
	    query.prepare("DELETE FROM cookies WHERE HEX(domain) = HEX(?) "
			  "AND passphrase_hash = ?");
	    query.bindValue
	      (0, dmisc::encodedString(domain.toUtf8(),
				       true).toBase64().constData());
	    query.bindValue(1, dmisc::passphraseHash(hashType));
	    query.exec();
	    query.prepare("DELETE FROM favorites WHERE HEX(domain) = HEX(?) "
			  "AND passphrase_hash = ?");
	    query.bindValue
	      (0, dmisc::encodedString(domain.toUtf8(),
				       true).toBase64().constData());
	    query.bindValue(1, dmisc::passphraseHash(hashType));
	    query.exec();
	  }

	db.close();
      }

      QSqlDatabase::removeDatabase("cookies");
    }
}

void dcookies::removeDomains(const QStringList &domains)
{
  QList<QNetworkCookie> all(allCookies());

  for(int i = all.size() - 1; i >= 0; i--)
    if(domains.contains(all.at(i).domain()))
      /*
      ** Remove all cookies from the specified domain.
      */

      all.removeAt(i);

  setAllCookies(all);
  all.clear();

  for(int i = 0; i < domains.size(); i++)
    m_favorites.remove(domains.at(i));

  if(dmisc::passphraseWasAuthenticated())
    {
      {
	QSqlDatabase db = QSqlDatabase::addDatabase("QSQLITE", "cookies");

	db.setDatabaseName(dooble::s_homePath +
			   QDir::separator() + "cookies.db");

	if(db.open())
	  {
	    QString hashType
	      (dooble::s_settings.value("settingsWindow/hashType").
	       toString());
	    QSqlQuery query(db);

	    query.exec("PRAGMA synchronous = OFF");

	    for(int i = 0; i < domains.size(); i++)
	      {
		query.prepare("DELETE FROM cookies WHERE HEX(domain) = HEX(?) "
			      "AND passphrase_hash = ?");
		query.bindValue
		  (0, dmisc::encodedString(domains.at(i).toUtf8(),
					   true).toBase64().constData());
		query.bindValue(1, dmisc::passphraseHash(hashType));
		query.exec();
		query.prepare
		  ("DELETE FROM favorites WHERE HEX(domain) = HEX(?) "
		   "AND passphrase_hash = ?");
		query.bindValue
		  (0, dmisc::encodedString(domains.at(i).toUtf8(),
					   true).toBase64().constData());
		query.bindValue(1, dmisc::passphraseHash(hashType));
		query.exec();
	      }
	  }

	db.close();
      }

      QSqlDatabase::removeDatabase("cookies");
    }
}

void dcookies::removeCookie(const QNetworkCookie &cookie)
{
  QList<QNetworkCookie> all(allCookies());

  for(int i = all.size() - 1; i >= 0; i--)
    if(all.at(i).domain() == cookie.domain() &&
       all.at(i).name() == cookie.name() &&
       all.at(i).path() == cookie.path())
      {
	/*
	** Remove exactly one cookie.
	*/

	deleteCookie(all.at(i));
	all.removeAt(i);
	break;
      }

  setAllCookies(all);
  all.clear();
}

void dcookies::removeCookie(const QString &name, const QString &domain,
			    const QString &path)
{
  QNetworkCookie cookie;

  cookie.setDomain(domain);
  cookie.setName(QVariant(name).toByteArray());
  cookie.setPath(path);
  removeCookie(cookie);
}

void dcookies::deleteCookie(const QNetworkCookie &cookie)
{
  if(dmisc::passphraseWasAuthenticated())
    {
      {
	QSqlDatabase db = QSqlDatabase::addDatabase("QSQLITE", "cookies");

	db.setDatabaseName(dooble::s_homePath +
			   QDir::separator() + "cookies.db");

	if(db.open())
	  {
	    QString hashType
	      (dooble::s_settings.value("settingsWindow/hashType").
	       toString());
	    QSqlQuery query(db);

	    query.exec("PRAGMA synchronous = OFF");
	    query.prepare
	      ("DELETE FROM cookies WHERE "
	       "HEX(domain) = HEX(?) AND "
	       "HEX(name) = HEX(?) AND "
	       "HEX(path) = HEX(?) AND "
	       "passphrase_hash = ?");
	    query.bindValue
	      (0, dmisc::encodedString(cookie.domain().toUtf8(),
				       true).toBase64().constData());
	    query.bindValue
	      (1, dmisc::encodedString(QVariant(cookie.name()).toString().
				       toUtf8(),
				       true).toBase64().constData());
	    query.bindValue
	      (2, dmisc::encodedString(cookie.path().toUtf8(),
				       true).toBase64().constData());
	    query.bindValue
	      (3, dmisc::passphraseHash(hashType));
	    query.exec();
	  }

	db.close();
      }

      QSqlDatabase::removeDatabase("cookies");
    }
}

void dcookies::createCookieDatabase(void)
{
  if(!dooble::s_settings.value("settingsWindow/cookiesEnabled",
			       true).toBool())
    return;

  {
    QSqlDatabase db = QSqlDatabase::addDatabase("QSQLITE", "cookies");

    db.setDatabaseName(dooble::s_homePath +
		       QDir::separator() + "cookies.db");

    if(db.open())
      {
	QSqlQuery query(db);

	query.exec("CREATE TABLE IF NOT EXISTS cookies ("
		   "domain TEXT NOT NULL, "
		   "name TEXT NOT NULL, "
		   "path TEXT NOT NULL, "
		   "expiration_date TEXT NOT NULL, "
		   "http TEXT NOT NULL, "
		   "secure TEXT NOT NULL, "
		   "value TEXT NOT NULL, "
		   "passphrase_hash TEXT NOT NULL, "
		   "PRIMARY KEY (domain, name, path, passphrase_hash))");
	query.exec("CREATE TABLE IF NOT EXISTS favorites ("
		   "domain TEXT NOT NULL, "
		   "passphrase_hash TEXT NOT NULL, "
		   "PRIMARY KEY (domain, passphrase_hash))");
      }

    db.close();
  }

  QSqlDatabase::removeDatabase("cookies");
}

void dcookies::allowDomain(const QString &domain,
			   const bool allowed)
{
  if(allowed)
    m_favorites[domain] = true;
  else
    m_favorites.remove(domain);

  if(dmisc::passphraseWasAuthenticated())
    {
      createCookieDatabase();

      {
	QSqlDatabase db = QSqlDatabase::addDatabase("QSQLITE", "cookies");

	db.setDatabaseName(dooble::s_homePath +
			   QDir::separator() + "cookies.db");

	if(db.open())
	  {
	    QString hashType
	      (dooble::s_settings.value("settingsWindow/hashType").
	       toString());
	    QSqlQuery query(db);

	    if(allowed)
	      {
		query.exec("PRAGMA synchronous = OFF");
		query.prepare("INSERT OR REPLACE INTO favorites "
			      "(domain, passphrase_hash)"
			      "VALUES (?, ?)");
		query.bindValue
		  (0, dmisc::encodedString(domain.toUtf8(),
					   true).toBase64());
		query.bindValue(1, dmisc::passphraseHash(hashType));
		query.exec();
	      }
	    else
	      {
		query.exec("PRAGMA synchronous = OFF");
		query.prepare("DELETE FROM favorites WHERE HEX(domain) = "
			      "HEX(?) AND "
			      "passphrase_hash = ?");
		query.bindValue
		  (0, dmisc::encodedString(domain.toUtf8(),
					   true).toBase64().constData());
		query.bindValue(1, dmisc::passphraseHash(hashType));
		query.exec();
	      }
	  }

	db.close();
      }

      QSqlDatabase::removeDatabase("cookies");
    }
}

void dcookies::populate(void)
{
  m_favorites.clear();
  deleteExpiredCookies();
  setAllCookies(QList<QNetworkCookie> ());

  if(!dmisc::passphraseWasAuthenticated())
    return;

  if(!m_deleteAndSaveCookiesTimer)
    {
      m_deleteAndSaveCookiesTimer = new QTimer(this);
      connect(m_deleteAndSaveCookiesTimer,
	      SIGNAL(timeout(void)),
	      this,
	      SLOT(slotDeleteAndSaveCookies(void)));
      m_deleteAndSaveCookiesTimer->setInterval(2500);
    }

  {
    QSqlDatabase db = QSqlDatabase::addDatabase("QSQLITE", "cookies");

    db.setDatabaseName(dooble::s_homePath +
		       QDir::separator() + "cookies.db");

    if(db.open())
      {
	QString hashType(dooble::s_settings.value("settingsWindow/hashType").
			 toString());
	QSqlQuery query(db);
	QList<QNetworkCookie> all;

	if(query.exec(QString("SELECT domain, name, path, expiration_date, "
			      "http, secure, value "
			      "FROM cookies WHERE passphrase_hash = '%1'").
		      arg(dmisc::passphraseHash(hashType))))
	  while(query.next())
	    {
	      QNetworkCookie cookie;

	      cookie.setDomain
		(QString::fromUtf8
		 (dmisc::decodedString
		  (QByteArray::fromBase64
		   (query.value(0).toByteArray())).constData()));

	      if(!QUrl::fromEncoded(cookie.domain().toUtf8(),
				    QUrl::StrictMode).isValid())
		{
		  QSqlQuery deleteQuery(db);

		  deleteQuery.exec("PRAGMA synchronous = OFF");
		  deleteQuery.exec(QString("DELETE FROM cookies WHERE "
					   "HEX(domain) = HEX('%1') AND "
					   "passphrase_hash = '%2'").
				   arg(query.value(0).toString()).
				   arg(dmisc::passphraseHash(hashType)));
		  continue;
		}

	      cookie.setName
		(QVariant
		 (QString::fromUtf8
		  (dmisc::decodedString
		   (QByteArray::fromBase64
		    (query.value(1).toByteArray())).constData())).
		 toByteArray());
	      cookie.setPath
		(QString::fromUtf8
		 (dmisc::decodedString
		  (QByteArray::fromBase64
		   (query.value(2).toByteArray())).constData()));
	      cookie.setExpirationDate
		(QDateTime::fromString
		 (QString::fromUtf8
		  (dmisc::decodedString
		   (QByteArray::fromBase64
		    (query.value(3).toByteArray())).constData()),
		  Qt::ISODate));
	      cookie.setHttpOnly
		(dmisc::decodedString
		 (QByteArray::fromBase64
		  (query.value(4).toByteArray())).toInt());
	      cookie.setSecure
		(dmisc::decodedString
		 (QByteArray::fromBase64
		  (query.value(5).toByteArray())).toInt());
	      cookie.setValue
		(QVariant
		 (QString::fromUtf8
		  (dmisc::decodedString
		   (QByteArray::fromBase64
		    (query.value(6).toByteArray())).constData())).
		 toByteArray());
	      all.append(cookie);
	    }

	if(!all.isEmpty())
	  setAllCookies(all);

	all.clear();

	if(query.exec(QString("SELECT domain FROM favorites WHERE "
			      "passphrase_hash = '%1'").
		      arg(dmisc::passphraseHash(hashType))))
	  while(query.next())
	    {
	      QNetworkCookie cookie;

	      cookie.setDomain
		(QString::fromUtf8
		 (dmisc::decodedString
		  (QByteArray::fromBase64
		   (query.value(0).toByteArray())).constData()));

	      if(!QUrl::fromEncoded(cookie.domain().toUtf8(),
				    QUrl::StrictMode).isValid())
		{
		  QSqlQuery deleteQuery(db);

		  deleteQuery.exec("PRAGMA synchronous = OFF");
		  deleteQuery.exec(QString("DELETE FROM favorites WHERE "
					   "HEX(domain) = HEX('%1') AND "
					   "passphrase_hash = '%2'").
				   arg(query.value(0).toString()).
				   arg(dmisc::passphraseHash(hashType)));
		  continue;
		}

	      m_favorites[cookie.domain()] = true;
	    }
      }

    db.close();
  }

  QSqlDatabase::removeDatabase("cookies");
}

void dcookies::saveCookie(const QNetworkCookie &cookie)
{
  if(cookie.isSessionCookie())
    return;
  else if(!dmisc::passphraseWasAuthenticated())
    return;

  if(dooble::s_settings.value("settingsWindow/cookiesShouldBe",
			      0).toInt() == 2)
    {
      QDateTime now(QDateTime::currentDateTime());

      if(cookie.expirationDate().toLocalTime() <= now)
	return;
    }

  createCookieDatabase();

  {
    QSqlDatabase db = QSqlDatabase::addDatabase("QSQLITE", "cookies");

    db.setDatabaseName(dooble::s_homePath +
		       QDir::separator() + "cookies.db");

    if(db.open())
      {
	QSqlQuery query(db);

	query.exec("PRAGMA synchronous = OFF");
	query.prepare("INSERT OR REPLACE INTO cookies "
		      "(domain, name, path, "
		      "expiration_date, http, secure, "
		      "value, passphrase_hash) "
		      "VALUES (?, ?, ?, ?, ?, ?, ?, ?)");
	query.bindValue
	  (0,
	   dmisc::encodedString(cookie.domain().toUtf8(), true).
	   toBase64());
	query.bindValue
	  (1,
	   dmisc::encodedString(QVariant(cookie.name()).toString().
				toUtf8(), true).toBase64());
	query.bindValue
	  (2,
	   dmisc::encodedString(cookie.path().toUtf8(), true).
	   toBase64());
	query.bindValue
	  (3,
	   dmisc::encodedString(cookie.expirationDate().
				toLocalTime().
				toString(Qt::ISODate).toUtf8(),
				true).toBase64());
	query.bindValue
	  (4,
	   dmisc::encodedString
	   (QString::number(cookie.isHttpOnly()).toAscii(),
	    true).toBase64());
	query.bindValue
	  (5,
	   dmisc::encodedString
	   (QString::number(cookie.isSecure()).toAscii(),
	    true).toBase64());
	query.bindValue
	  (6,
	   dmisc::encodedString(QVariant(cookie.value()).toString().
				toUtf8(), true).toBase64());

	QString hashType(dooble::s_settings.value("settingsWindow/hashType").
			 toString());

	query.bindValue(7, dmisc::passphraseHash(hashType));
	query.exec();
      }

    db.close();
  }

  QSqlDatabase::removeDatabase("cookies");
}

void dcookies::deleteExpiredCookies(void)
{
  if(!dmisc::passphraseWasAuthenticated())
    return;
  else if(dooble::s_settings.value("settingsWindow/cookiesShouldBe",
				   0).toInt() != 2)
    return;

  {
    QSqlDatabase db = QSqlDatabase::addDatabase("QSQLITE", "cookies");

    db.setDatabaseName(dooble::s_homePath +
		       QDir::separator() + "cookies.db");

    if(db.open())
      {
	QString hashType(dooble::s_settings.value("settingsWindow/hashType").
			 toString());
	QSqlQuery query(db);

	if(query.exec(QString("SELECT expiration_date FROM cookies "
			      "WHERE passphrase_hash = '%1'").
		      arg(dmisc::passphraseHash(hashType))))
	  {
	    QDateTime now(QDateTime::currentDateTime());
	    QSqlQuery deleteQuery(db);

	    deleteQuery.exec("PRAGMA synchronous = OFF");

	    while(query.next())
	      {
		QDateTime dateTime
		  (QDateTime::fromString
		   (QString::fromUtf8
		    (dmisc::decodedString
		     (QByteArray::fromBase64
		      (query.value(0).toByteArray())).constData()),
		    Qt::ISODate));

		if(dateTime.isValid())
		  {
		    if(dateTime.toLocalTime() <= now)
		      deleteQuery.exec(QString("DELETE FROM cookies WHERE "
					       "HEX(expiration_date) = "
					       "HEX('%1') AND "
					       "passphrase_hash = '%2'").arg
				       (query.value(0).toString()).arg
				       (dmisc::passphraseHash(hashType)));
		  }
		else
		  deleteQuery.exec(QString("DELETE FROM cookies WHERE "
					   "HEX(expiration_date) = "
					   "HEX('%1') AND "
					   "passphrase_hash = '%2'").arg
				   (query.value(0).toString()).arg
				   (dmisc::passphraseHash(hashType)));
	      }
	  }
      }

    db.close();
  }

  QSqlDatabase::removeDatabase("cookies");
}

int dcookies::computedTimerInterval(void) const
{
  int interval = 0;

  if(dooble::s_settings.value("settingsWindow/cookiesEnabled",
			      true).toBool() &&
     dooble::s_settings.value("settingsWindow/cookieTimerEnabled",
			      false).toBool())
    {
      int unit = dooble::s_settings.value
	("settingsWindow/cookieTimerUnit", 0).toInt();

      if(!(unit == 0 || unit == 1))
	unit = 0;

      interval = dooble::s_settings.value
	("settingsWindow/cookieTimerInterval", 1).toInt();

      if(unit == 0)
	{
	  if(interval < 1 || interval > 60)
	    interval = 1;

	  interval = 3600000 * interval;
	}
      else
	{
	  if(interval < 1 || interval > 3600)
	    interval = 1;

	  interval = 60000 * interval;
	}
    }

  return interval;
}

void dcookies::slotCookieTimerChanged(void)
{
  int interval = computedTimerInterval();

  if(interval > 0)
    {
      if(interval != m_timer->interval())
	m_timer->start(interval);
    }
  else
    m_timer->stop();
}

bool dcookies::isFavorite(const QString &domain) const
{
  if(m_favorites.contains(domain))
    return m_favorites.value(domain);
  else
    return false;
}

void dcookies::slotRemoveDomains(void)
{
  QStringList domains;
  QList<QNetworkCookie> all(allCookies());

  /*
  ** Locate the domains that are not in the
  ** m_favorites container.
  */

  for(int i = 0; i < all.size(); i++)
    if(m_favorites.contains(all.at(i).domain()))
      {
	if(!m_favorites.value(all.at(i).domain()))
	  if(!domains.contains(all.at(i).domain()))
	    domains.append(all.at(i).domain());
      }
    else if(!domains.contains(all.at(i).domain()))
      domains.append(all.at(i).domain());

  all.clear();

  /*
  ** Remove all cookies that are located in the
  ** domains container.
  */

  for(int i = 0; i < domains.size(); i++)
    removeDomain(domains.at(i));

  emit domainsRemoved(domains);
}

void dcookies::reencode(QProgressBar *progress)
{
  /*
  ** This method assumes that the respective container
  ** is as current as possible. It will use the container's
  ** contents to create new database.
  ** Obsolete data will be deleted via the purge method.
  */

  if(progress)
    {
      progress->setMaximum(-1);
      progress->setVisible(true);
    }

  QHash<QString, bool> favorites(m_favorites);
  QList<QNetworkCookie> all(allCookies());

  if(progress)
    {
      progress->setMaximum(all.size());
      progress->update();
    }

  for(int i = 0; i < all.size(); i++)
    {
      saveCookie(all.at(i));

      if(progress)
	progress->setValue(i + 1);
    }

  all.clear();
  m_favorites.clear();

  if(progress)
    {
      progress->setMaximum(favorites.size());
      progress->update();
    }

  for(int i = 0; i < favorites.keys().size(); i++)
    {
      allowDomain(favorites.keys().at(i),
		  favorites[favorites.keys().at(i)]);

      if(progress)
	progress->setValue(i + 1);
    }

  favorites.clear();
  purge();

  if(progress)
    progress->setVisible(false);
}

void dcookies::purge(void)
{
  {
    QSqlDatabase db = QSqlDatabase::addDatabase("QSQLITE", "cookies");

    db.setDatabaseName(dooble::s_homePath +
		       QDir::separator() + "cookies.db");

    if(db.open())
      {
	QString hashType(dooble::s_settings.value
			 ("settingsWindow/hashType").
			 toString());
	QSqlQuery query(db);

	query.exec("PRAGMA synchronous = OFF");

	if(dmisc::passphraseWasAuthenticated())
	  {
	    query.exec
	      (QString("DELETE FROM cookies WHERE "
		       "passphrase_hash <> '%1'").
	       arg(dmisc::passphraseHash(hashType)));
	    query.exec
	      (QString("DELETE FROM favorites WHERE "
		       "passphrase_hash <> '%1'").
	       arg(dmisc::passphraseHash(hashType)));
	  }
	else
	  {
	    query.exec
	      (QString("DELETE FROM cookies WHERE "
		       "passphrase_hash = '%1'").
	       arg(dmisc::passphraseHash(hashType)));
	    query.exec
	      (QString("DELETE FROM favorites WHERE "
		       "passphrase_hash = '%1'").
	       arg(dmisc::passphraseHash(hashType)));
	  }

	db.close();
      }
  }

  QSqlDatabase::removeDatabase("cookies");
}

QList<QNetworkCookie> dcookies::allCookiesAndFavorites(void) const
{
  QList<QNetworkCookie> favorites;

  for(int i = 0; i < m_favorites.keys().size(); i++)
    {
      QNetworkCookie cookie;

      cookie.setDomain(m_favorites.keys().at(i));
      favorites.append(cookie);
    }

  return allCookies() + favorites;
}

void dcookies::slotDeleteAndSaveCookies(void)
{
  if(!dmisc::passphraseWasAuthenticated())
    return;

  {
    QSqlDatabase db = QSqlDatabase::addDatabase("QSQLITE", "cookies");

    db.setDatabaseName(dooble::s_homePath + QDir::separator() + "cookies.db");

    if(db.open())
      {
	QString hashType
	  (dooble::s_settings.value("settingsWindow/hashType").toString());
	QDateTime now(QDateTime::currentDateTime());
	QSqlQuery query(db);

	query.exec("PRAGMA synchronous = OFF");
	query.prepare
	  ("DELETE FROM cookies WHERE "
	   "HEX(domain) = HEX(?) AND "
	   "HEX(name) = HEX(?) AND "
	   "HEX(path) = HEX(?) AND "
	   "passphrase_hash = ?");

	while(!m_deleteCookies.isEmpty())
	  {
	    if(QApplication::hasPendingEvents())
	      break;

	    QNetworkCookie cookie(m_deleteCookies.takeFirst());

	    query.bindValue
	      (0, dmisc::encodedString(cookie.domain().toUtf8(),
				       true).toBase64().constData());
	    query.bindValue
	      (1, dmisc::encodedString(QVariant(cookie.name()).
				       toString().toUtf8(),
				       true).toBase64().constData());
	    query.bindValue
	      (2, dmisc::encodedString(cookie.path().toUtf8(),
				       true).toBase64().constData());
	    query.bindValue
	      (3, dmisc::passphraseHash(hashType));
	    query.exec();
	  }

	query.prepare("INSERT OR REPLACE INTO cookies "
		      "(domain, name, path, "
		      "expiration_date, http, secure, "
		      "value, passphrase_hash) "
		      "VALUES (?, ?, ?, ?, ?, ?, ?, ?)");

	while(!m_saveCookies.isEmpty())
	  {
	    if(QApplication::hasPendingEvents())
	      break;

	    QNetworkCookie cookie(m_saveCookies.takeFirst());

	    if(cookie.isSessionCookie())
	      continue;
	    else if(dooble::s_settings.value("settingsWindow/cookiesShouldBe",
					     0).toInt() == 2)
	      {
		if(cookie.expirationDate().toLocalTime() <= now)
		  continue;
	      }

	    query.bindValue
	      (0,
	       dmisc::encodedString(cookie.domain().toUtf8(),
				    true).toBase64());
	    query.bindValue
	      (1,
	       dmisc::encodedString(QVariant(cookie.name()).
				    toString().toUtf8(), true).toBase64());
	    query.bindValue
	      (2,
	       dmisc::encodedString(cookie.path().toUtf8(), true).
	       toBase64());
	    query.bindValue
	      (3,
	       dmisc::encodedString(cookie.expirationDate().
				    toLocalTime().
				    toString(Qt::ISODate).toUtf8(),
				    true).toBase64());
	    query.bindValue
	      (4,
	       dmisc::encodedString
	       (QString::number(cookie.isHttpOnly()).toAscii(),
		true).toBase64());
	    query.bindValue
	      (5,
	       dmisc::encodedString
	       (QString::number(cookie.isSecure()).toAscii(),
		true).toBase64());
	    query.bindValue
	      (6,
	       dmisc::encodedString(QVariant(cookie.value()).
				    toString().toUtf8(), true).toBase64());
	    query.bindValue(7, dmisc::passphraseHash(hashType));
	    query.exec();
	  }
      }

    db.close();
  }

  QSqlDatabase::removeDatabase("cookies");

  if(m_deleteCookies.isEmpty() && m_saveCookies.isEmpty())
    if(m_deleteAndSaveCookiesTimer)
      m_deleteAndSaveCookiesTimer->stop();
}
