/****************************************************************************
** Dooble - The Secure Internet Web Browser
**
** Copyright (c) 2008, 2009, 2010, 2011 Alexis Megas,
** Gunther van Dooble, and the Dooble Team.
** All rights reserved.
**
** License: GPL2 only:
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; version 2 of the License only.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
** or see here: http://www.gnu.org/licenses/gpl.html
**
** For the WebKit library, please see: http://webkit.org.
**
** THE CODE IS PROVIDED BY THE AUTHORS ''AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
** IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
** PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
** ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
** DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
** GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
** INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
** IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
** ARISING IN ANY WAY OUT OF THE USE OF THIS APPLICATION, EVEN IF ADVISED
** OF THE POSSIBILITY OF SUCH DAMAGE.
**
** Please report all praise, requests, bugs, and problems to the project
** team and administrators: http://sf.net/projects/dooble.
**
** You can find us listed at our project page. New team members are welcome.
** The name of the authors should not be used to endorse or promote products
** derived from Dooble without specific prior written permission.
** If you use this code for other projects, please let us know.
**
** Web sites:
**   http://sf.net/projects/dooble
**   http://dooble.sf.net
****************************************************************************/

#include <QDir>
#include <QUrl>
#include <QIcon>
#include <QMenu>
#include <QTimer>
#include <QBuffer>
#include <QKeyEvent>
#include <QSettings>
#include <QSqlQuery>
#include <QClipboard>
#include <QScrollBar>
#include <QSqlRecord>
#include <QProgressBar>

#include "dmisc.h"
#include "dooble.h"
#include "dhistory.h"

/*
** m_contentsList is used primarily for re-encoding.
** One cannot rely on the model since it may not
** represent the user's complete history.
*/

dhistory::dhistory(void):QMainWindow()
{
  ui.setupUi(this);
#ifdef Q_WS_MAC
  setAttribute(Qt::WA_MacMetalStyle, false);
  statusBar()->setSizeGripEnabled(false);
#endif
  m_model = new QStandardItemModel(this);
  m_model->setSortRole(Qt::UserRole);

  QStringList list;

  list << tr("Title") << tr("Host")
       << tr("Location") << tr("Last Time Visited")
       << tr("Visits") << "Hash" << "Description";
  m_model->setHorizontalHeaderLabels(list);
  list.clear();
  ui.history->setModel(m_model);
  ui.history->setColumnHidden(m_model->columnCount() - 1, true);
  ui.history->setColumnHidden(m_model->columnCount() - 2, true);
  ui.history->setSelectionMode(QAbstractItemView::ExtendedSelection);
  ui.history->setSortingEnabled(true);
  ui.history->horizontalHeader()->setDefaultAlignment
    (Qt::AlignLeft);
  ui.history->horizontalHeader()->setSortIndicator(0, Qt::AscendingOrder);
  ui.history->horizontalHeader()->setSortIndicatorShown(true);
  ui.history->horizontalHeader()->setStretchLastSection(true);

  for(int i = 0; i < ui.history->horizontalHeader()->count() - 2; i++)
    ui.history->resizeColumnToContents(i);

  ui.history->horizontalHeader()->setResizeMode(QHeaderView::Interactive);
  ui.searchLineEdit->setPlaceholderText(tr("Search History"));
  slotSetIcons();
  connect(ui.history->horizontalHeader(),
	  SIGNAL(sortIndicatorChanged(int, Qt::SortOrder)),
	  this,
	  SLOT(slotSort(int, Qt::SortOrder)));
  connect(ui.deletePushButton,
	  SIGNAL(clicked(void)),
	  this,
	  SLOT(slotDeletePage(void)));
  connect(ui.deleteAllPushButton,
	  SIGNAL(clicked(void)),
	  this,
	  SLOT(slotDeleteAll(void)));
  connect(ui.treeWidget,
	  SIGNAL(itemSelectionChanged(void)),
	  this,
	  SLOT(slotItemSelectionChanged(void)));
  connect(this,
	  SIGNAL(iconsChanged(void)),
	  ui.searchLineEdit,
	  SLOT(slotSetIcons(void)));
  connect(ui.searchLineEdit, SIGNAL(textEdited(const QString &)),
	  this, SLOT(slotTextChanged(const QString &)));
  connect(ui.history,
	  SIGNAL(doubleClicked(const QModelIndex &)),
	  this,
	  SLOT(slotItemDoubleClicked(const QModelIndex &)));
  connect(ui.closePushButton,
	  SIGNAL(clicked(void)),
	  this,
	  SLOT(close(void)));
  connect(ui.history->selectionModel(),
	  SIGNAL(selectionChanged(const QItemSelection &,
				  const QItemSelection &)),
	  this,
	  SLOT(slotItemsSelected(const QItemSelection &,
				 const QItemSelection &)));
  ui.history->setContextMenuPolicy(Qt::CustomContextMenu);
  connect(ui.history, SIGNAL(customContextMenuRequested(const QPoint &)),
	  SLOT(slotShowContextMenu(const QPoint &)));

  m_timer = new QTimer(this);
  m_timer->setInterval(2500);
  connect(m_timer,
	  SIGNAL(timeout(void)),
	  this,
	  SLOT(slotTimeout(void)));
  ui.splitter->setStretchFactor(0, 0);
  ui.splitter->setStretchFactor(1, 1);

  QTreeWidgetItem *item1 = 0;
  QList<QTreeWidgetItem *> items;

  item1 = new QTreeWidgetItem((QTreeWidget *) 0,
			      QStringList(tr("All")));
  item1->setData(0, Qt::UserRole, "");
  items << item1;
  item1 = new QTreeWidgetItem((QTreeWidget *) 0,
			      QStringList(tr("Today")));
  item1->setData
    (0, Qt::UserRole, QDate::currentDate().toString(Qt::ISODate));
  items << item1;
  item1 = new QTreeWidgetItem((QTreeWidget *) 0,
			      QStringList(tr("Yesterday")));
  item1->setData
    (0, Qt::UserRole, QDate::currentDate().addDays(-1).
     toString(Qt::ISODate));
  items << item1;

  QDate currentMonth(QDate::currentDate());
  QTreeWidgetItem *item2 = new QTreeWidgetItem
    ((QTreeWidget *) 0,
     QStringList(currentMonth.toString("MMMM").toUtf8().constData()));

  item2->setData
    (0, Qt::UserRole, currentMonth.toString(Qt::ISODate).mid(0, 7));
  items << item2;

  QDate lastMonth(QDate::currentDate().addMonths(-1));
  QTreeWidgetItem *item3 = new QTreeWidgetItem
    ((QTreeWidget *) 0,
     QStringList(tr(lastMonth.toString("MMMM").toUtf8().constData())));

  item3->setData
    (0, Qt::UserRole, lastMonth.toString(Qt::ISODate).mid(0, 7));
  items << item3;
  ui.treeWidget->insertTopLevelItems(0, items);

  for(int i = 0; i < currentMonth.daysInMonth(); i++)
    {
      QDate date;

      date.setDate(currentMonth.year(), currentMonth.month(), i + 1);
      item1 = new QTreeWidgetItem
	(item2, QStringList(tr(QString::number(i + 1).toAscii().
			       constData())));
      item1->setData(0, Qt::UserRole, date.toString(Qt::ISODate));
    }

  for(int i = 0; i < lastMonth.daysInMonth(); i++)
    {
      QDate date;

      date.setDate(lastMonth.year(), lastMonth.month(), i + 1);
      item1 = new QTreeWidgetItem
	(item3, QStringList(tr(QString::number(i + 1).toAscii().constData())));
      item1->setData(0, Qt::UserRole, date.toString(Qt::ISODate));
    }

  items.at(1)->setSelected(true);
  ui.treeWidget->setFocus();

  if(dooble::s_settings.contains("historyWindow/splitterState"))
    ui.splitter->restoreState
      (dooble::s_settings.value("historyWindow/splitterState",
				"").toByteArray());

  if(dooble::s_settings.contains("historyWindow/tableColumnsState5"))
    if(!ui.history->horizontalHeader()->restoreState
       (dooble::s_settings.value("historyWindow/tableColumnsState5",
				 "").toByteArray()))
      {
	ui.history->horizontalHeader()->setDefaultAlignment
	  (Qt::AlignLeft);
	ui.history->horizontalHeader()->setSortIndicator(0,
							 Qt::AscendingOrder);
	ui.history->horizontalHeader()->setSortIndicatorShown(true);
	ui.history->horizontalHeader()->setStretchLastSection(true);
	ui.history->horizontalHeader()->setResizeMode
	  (QHeaderView::Interactive);
      }

  ui.history->horizontalHeader()->setMovable(true);
}

dhistory::~dhistory()
{
  saveState();

  if(!dmisc::passphraseWasAuthenticated())
    {
      {
	QSqlDatabase db = QSqlDatabase::addDatabase("QSQLITE", "history");

	db.setDatabaseName(dooble::s_homePath +
			   QDir::separator() + "history.db");

	if(db.open())
	  {
	    QString hashType(dooble::s_settings.value
			     ("settingsWindow/hashType").
			     toString());
	    QSqlQuery query(db);

	    query.exec("PRAGMA synchronous = OFF");
	    query.exec
	      (QString("DELETE FROM history WHERE "
		       "passphrase_hash = '%1'").
	       arg(dmisc::passphraseHash(hashType)));
	  }

	db.close();
      }

      QSqlDatabase::removeDatabase("history");
    }
  else
    purge();
}

void dhistory::show(QWidget *parent)
{
  m_parent = parent;
  disconnect(m_parent,
	     SIGNAL(destroyed(void)),
	     this,
	     SLOT(slotParentDestroyed(void)));
  connect(m_parent,
	  SIGNAL(destroyed(void)),
	  this,
	  SLOT(slotParentDestroyed(void)));
  disconnect(this, SIGNAL(open(const QUrl &)), 0, 0);
  disconnect(this, SIGNAL(createTab(const QUrl &)), 0, 0);
  disconnect(this, SIGNAL(openInNewWindow(const QUrl &)), 0, 0);
  connect(this, SIGNAL(open(const QUrl &)),
	  m_parent, SLOT(slotLoadPage(const QUrl &)));
  connect(this, SIGNAL(createTab(const QUrl &)),
	  m_parent, SLOT(slotOpenLinkInNewTab(const QUrl &)));
  connect(this, SIGNAL(openInNewWindow(const QUrl &)),
	  m_parent, SLOT(slotOpenLinkInNewWindow(const QUrl &)));

  if(!isVisible())
    populate();

  if(!isVisible())
   {
     /*
     ** Don't annoy the user.
     */

     if(dooble::s_settings.contains("historyWindow/geometry"))
       {
	 QByteArray g(dooble::s_settings.value("historyWindow/geometry").
		      toByteArray());

	 if(!restoreGeometry(g))
	   setGeometry(100, 100, 800, 600);
       }
     else
       setGeometry(100, 100, 800, 600);
   }

  showNormal();
  raise();

  if(!m_timer->isActive())
    m_timer->start();
}

void dhistory::purge(void)
{
  {
    QSqlDatabase db = QSqlDatabase::addDatabase
      ("QSQLITE", "history_window_purge_all");

    db.setDatabaseName(dooble::s_homePath +
		       QDir::separator() + "history.db");

    if(db.open())
      {
	/*
	** Delete expired history items and items having invalid
	** URLs.
	*/

	QString hashType(dooble::s_settings.value("settingsWindow/hashType").
			 toString());
	QSqlQuery query(db);

	if(query.exec(QString("SELECT url, last_visited FROM history "
			      "WHERE passphrase_hash = '%1'").
		      arg(dmisc::passphraseHash(hashType))))
	  {
	    QDate date(QDate::currentDate());
	    QSqlQuery deleteQuery(db);

	    date = date.addDays
	      (-dooble::s_settings.value("settingsWindow/historyDays",
					 8).toInt());
	    deleteQuery.exec
	      ("PRAGMA synchronous = OFF");

	    while(query.next())
	      {
	        QUrl url
		  (QUrl::fromEncoded
		   (dmisc::decodedString
		    (QByteArray::fromBase64
		     (query.value(0).toByteArray())).constData(),
		    QUrl::StrictMode));

		if(url.isValid())
		  {
		    QDate urlDate
		      (QDate::fromString
		       (QString::fromUtf8
			(dmisc::decodedString
			 (QByteArray::fromBase64
			  (query.value(1).toByteArray())).constData()),
			Qt::ISODate));

		    if(urlDate <= date)
		      deleteQuery.exec
			(QString("DELETE FROM history WHERE "
				 "HEX(url) = HEX('%1') AND "
				 "passphrase_hash = '%2'").
			 arg(query.value(0).toString()).
			 arg(dmisc::passphraseHash(hashType)));
		  }
		else
		  deleteQuery.exec
		    (QString("DELETE FROM history WHERE "
			     "HEX(url) = HEX('%1') AND "
			     "passphrase_hash = '%2'").
		     arg(query.value(0).toString()).
		     arg(dmisc::passphraseHash(hashType)));
	      }
	  }

	if(dmisc::passphraseWasAuthenticated())
	  {
	    /*
	    ** Delete session-based history items.
	    */

	    QSqlQuery deleteQuery(db);

	    deleteQuery.exec
	      ("PRAGMA synchronous = OFF");
	    deleteQuery.exec
	      (QString("DELETE FROM history WHERE passphrase_hash <> "
		       "'%1'").arg(dmisc::passphraseHash(hashType)));
	  }
      }

    db.close();
  }

  QSqlDatabase::removeDatabase("history_window_purge_all");
  statusBar()->showMessage(QString(tr("%1 item(s) - %2 selected")).
			   arg(m_model->rowCount()).
			   arg(ui.history->selectionModel()->
			       selectedRows(0).size()));
}

void dhistory::slotSort(int column, Qt::SortOrder order)
{
  m_model->sort(column, order);
}

void dhistory::slotSetIcons(void)
{
  QSettings settings
    (dooble::s_settings.value("iconSet").toString(), QSettings::IniFormat);

  setWindowIcon
    (QIcon(settings.value("historyWindow/windowIcon").toString()));
  ui.closePushButton->setIcon
    (QIcon(settings.value("cancelButtonIcon").toString()));
  ui.deletePushButton->setIcon
    (QIcon(settings.value("historyWindow/deleteButtonIcon").toString()));
  ui.deleteAllPushButton->setIcon
    (QIcon(settings.value("historyWindow/deleteAllButtonIcon").toString()));
  emit iconsChanged();
}

void dhistory::slotTimeout(void)
{
  QFileInfo fileInfo(dooble::s_homePath + QDir::separator() + "history.db");
  static QDateTime lastModificationTime;

  if(fileInfo.exists())
    {
      if(fileInfo.lastModified() <= lastModificationTime)
	return;
      else
	lastModificationTime = fileInfo.lastModified();
    }
  else
    lastModificationTime = QDateTime();

  populate();
}

void dhistory::slotDeletePage(void)
{
  QApplication::setOverrideCursor(QCursor(Qt::WaitCursor));

  /*
  ** Delete an entry, or more, from the history table.
  */

  m_timer->stop();

  {
    QSqlDatabase db = QSqlDatabase::addDatabase
      ("QSQLITE", "history_window_delete");

    db.setDatabaseName(dooble::s_homePath +
		       QDir::separator() + "history.db");

    if(db.open())
      {
	int removedRows = 0;
	QString hashType(dooble::s_settings.value("settingsWindow/hashType").
			 toString());
	QSqlQuery query(db);
	QModelIndexList list(ui.history->selectionModel()->selectedRows(5));

	/*
	** Sort the list by row number. If the list is not sorted,
	** the following while-loop misbehaves.
	*/

	qSort(list);
	query.exec("PRAGMA synchronous = OFF");

	while(!list.isEmpty())
	  {
	    QStandardItem *item = m_model->item
	      (list.takeFirst().row() - removedRows, 5);

	    if(item)
	      {
		query.exec(QString("DELETE FROM history "
				   "WHERE url_hash = '%1' AND "
				   "passphrase_hash = '%2'").
			   arg(item->text()).
			   arg(dmisc::passphraseHash(hashType)));

		if(m_model->removeRow(item->row()))
		  removedRows += 1;
	      }
	  }
      }

    db.close();
  }

  QSqlDatabase::removeDatabase("history_window_delete");
  statusBar()->showMessage(QString(tr("%1 item(s) - %2 selected")).
			   arg(m_model->rowCount()).
			   arg(ui.history->selectionModel()->
			       selectedRows(0).size()));
  m_timer->start();
  QApplication::restoreOverrideCursor();
}

void dhistory::slotOpen(void)
{
  QStandardItem *item = 0;
  QModelIndexList list(ui.history->selectionModel()->selectedRows(2));

  if(!list.isEmpty())
    item = m_model->itemFromIndex(list.takeFirst());

  if(item)
    emit open(QUrl::fromUserInput(item->text()));

  list.clear();
}

void dhistory::slotCopyUrl(void)
{
  QStandardItem *item = 0;
  QModelIndexList list(ui.history->selectionModel()->selectedRows(2));

  if(!list.isEmpty())
    item = m_model->itemFromIndex(list.takeFirst());

  if(item && QApplication::clipboard())
    QApplication::clipboard()->setText(item->text());

  list.clear();
}

void dhistory::slotOpenInNewTab(void)
{
  QModelIndexList list(ui.history->selectionModel()->selectedRows(2));

  while(!list.isEmpty())
    {
      QUrl url;
      QStandardItem *item = m_model->itemFromIndex(list.takeFirst());

      if(item)
	emit createTab(QUrl::fromUserInput(item->text()));
    }
}

void dhistory::slotOpenInNewWindow(void)
{
  QStandardItem *item = 0;
  QModelIndexList list(ui.history->selectionModel()->selectedRows(2));

  if(!list.isEmpty())
    item = m_model->itemFromIndex(list.takeFirst());

  if(item)
    emit openInNewWindow(QUrl::fromUserInput(item->text()));

  list.clear();
}

void dhistory::slotShowContextMenu(const QPoint &point)
{
  if(!ui.history->selectionModel()->selectedRows().isEmpty())
    {
      QMenu menu(this);

      if(ui.history->selectionModel()->selectedRows().size() == 1 &&
	 ui.history->indexAt(point).isValid())
	{
	  menu.addAction(tr("&Bookmark"),
			 this, SLOT(slotBookmark(void)));
	  menu.addSeparator();
	  menu.addAction(tr("&Copy URL"),
			 this, SLOT(slotCopyUrl(void)));
	  menu.addSeparator();
	  menu.addAction(tr("&Delete Page"),
			 this, SLOT(slotDeletePage(void)));
	  menu.addSeparator();
	  menu.addAction(tr("Open in &Current Tab"),
			 this, SLOT(slotOpen(void)));
	  menu.addAction(tr("Open in New &Tab"),
			 this, SLOT(slotOpenInNewTab(void)));
	  menu.addAction(tr("Open in &New Window"),
			 this, SLOT(slotOpenInNewWindow(void)));
	}
      else
	{
	  menu.addAction(tr("&Bookmark"),
			 this, SLOT(slotBookmark(void)));
	  menu.addSeparator();
	  menu.addAction(tr("&Delete Pages"),
			 this, SLOT(slotDeletePage(void)));
	  menu.addSeparator();
	  menu.addAction(tr("Open in &New Tabs"),
			 this, SLOT(slotOpenInNewTab(void)));
	}

      menu.exec(ui.history->mapToGlobal(point));
    }
}

void dhistory::slotItemDoubleClicked(const QModelIndex &index)
{
  QStandardItem *item1 = 0;

  item1 = m_model->itemFromIndex(index);

  if(item1)
    {
      QStandardItem *item2 = m_model->item(item1->row(), 2);

      if(item2)
	emit open(QUrl::fromUserInput(item2->text()));
    }
}

void dhistory::slotParentDestroyed(void)
{
  /*
  ** Locate another Dooble window (parent).
  */

  foreach(QWidget *widget, QApplication::topLevelWidgets())
    if(qobject_cast<dooble *> (widget))
      if(widget != m_parent)
	{
	  m_parent = widget;
	  connect(m_parent,
		  SIGNAL(destroyed(void)),
		  this,
		  SLOT(slotParentDestroyed(void)));
	  disconnect(this, SIGNAL(open(const QUrl &)),
		     m_parent, SLOT(slotLoadPage(const QUrl &)));
	  disconnect(this, SIGNAL(createTab(const QUrl &)),
		     m_parent, SLOT(slotOpenLinkInNewTab(const QUrl &)));
	  disconnect(this, SIGNAL(openInNewWindow(const QUrl &)),
		     m_parent, SLOT(slotOpenLinkInNewWindow(const QUrl &)));
	  connect(this, SIGNAL(open(const QUrl &)),
		  m_parent, SLOT(slotLoadPage(const QUrl &)));
	  connect(this, SIGNAL(createTab(const QUrl &)),
		  m_parent, SLOT(slotOpenLinkInNewTab(const QUrl &)));
	  connect(this, SIGNAL(openInNewWindow(const QUrl &)),
		  m_parent, SLOT(slotOpenLinkInNewWindow(const QUrl &)));
	  break;
	}
}

void dhistory::slotDeleteAll(void)
{
  QApplication::setOverrideCursor(QCursor(Qt::WaitCursor));

  m_timer->stop();

  {
    QSqlDatabase db = QSqlDatabase::addDatabase
      ("QSQLITE", "history_window_delete_all");

    db.setDatabaseName(dooble::s_homePath +
		       QDir::separator() + "history.db");

    if(db.open())
      {
	QString hashType(dooble::s_settings.value("settingsWindow/hashType").
			 toString());
	QSqlQuery query(db);

	query.exec("PRAGMA synchronous = OFF");

	for(int i = m_model->rowCount() - 1; i >= 0; i--)
	  {
	    QStandardItem *item = m_model->item(i, 5);

	    if(item)
	      {
		query.exec(QString("DELETE FROM history "
				   "WHERE url_hash = '%1' AND "
				   "passphrase_hash = '%2'").
			   arg(item->text()).
			   arg(dmisc::passphraseHash(hashType)));
		m_model->removeRow(item->row());
	      }
	  }
      }

    db.close();
  }

  QSqlDatabase::removeDatabase("history_window_delete_all");
  statusBar()->showMessage(QString(tr("%1 item(s) - %2 selected")).
			   arg(m_model->rowCount()).
			   arg(ui.history->selectionModel()->
			       selectedRows(0).size()));
  m_timer->start();
  QApplication::restoreOverrideCursor();
}

void dhistory::populate(void)
{
  QApplication::setOverrideCursor(QCursor(Qt::WaitCursor));

  /*
  ** Remember the user's current selection. After the table
  ** is populated, highlight approximate selections.
  */

  int hPos = ui.history->horizontalScrollBar()->value();
  int vPos = ui.history->verticalScrollBar()->value();
  QModelIndexList list(ui.history->selectionModel()->selectedRows());

  {
    QSqlDatabase db = QSqlDatabase::addDatabase("QSQLITE", "history_window");

    db.setDatabaseName(dooble::s_homePath +
		       QDir::separator() + "history.db");

    if(db.open())
      {
	m_model->removeRows(0, m_model->rowCount());
	m_contentsList.clear();

	QString dateStr("");
	QString hashType(dooble::s_settings.value("settingsWindow/hashType").
			 toString());
	QString searchStr
	  (ui.searchLineEdit->text().toLower().trimmed());
	QSqlQuery query(db);
	QList<QTreeWidgetItem *> list(ui.treeWidget->selectedItems());

	if(!list.isEmpty())
	  {
	    QTreeWidgetItem *item = list.at(0);

	    dateStr = item->data(0, Qt::UserRole).toString();
	  }

	list.clear();

	if(query.exec(QString("SELECT title, host, url, last_visited, "
			      "visits, icon, url_hash, description "
			      "FROM history WHERE "
			      "passphrase_hash = '%1'").
		      arg(dmisc::passphraseHash(hashType))))
	  while(query.next())
	    {
	      QUrl url
		(QUrl::fromEncoded
		 (dmisc::decodedString
		  (QByteArray::fromBase64
		   (query.value(2).toByteArray())).constData(),
		  QUrl::StrictMode));

	      if(!dmisc::isSchemeAcceptedByDooble(url.scheme()))
		{
		  QSqlQuery deleteQuery(db);

		  deleteQuery.exec
		    ("PRAGMA synchronous = OFF");
		  deleteQuery.exec(QString("DELETE FROM history WHERE "
					   "HEX(url) = HEX('%1') AND "
					   "passphrase_hash = '%2'").
				   arg(query.value(2).toString()).
				   arg(dmisc::passphraseHash(hashType)));
		  continue;
		}

	      QString host
		(QString::fromUtf8
		 (dmisc::decodedString
		  (QByteArray::fromBase64
		   (query.value(1).toByteArray())).constData()));
	      QString title
		(QString::fromUtf8
		 (dmisc::decodedString
		  (QByteArray::fromBase64
		   (query.value(0).toByteArray())).constData()));
	      QString visits
		(dmisc::decodedString
		 (QByteArray::fromBase64
		  (query.value(4).toByteArray())).constData());
	      QString urlHash(query.value(6).toString());
	      QString description
		(QString::fromUtf8
		 (dmisc::decodedString
		  (QByteArray::fromBase64
		   (query.value(7).toByteArray())).constData()));
	      QDateTime dateTime
		(QDateTime::fromString
		 (QString::fromUtf8
		  (dmisc::decodedString
		   (QByteArray::fromBase64
		    (query.value(3).toByteArray())).constData()),
		  Qt::ISODate));
	      QByteArray bytes
		(dmisc::decodedString
		 (QByteArray::fromBase64
		  (query.value(5).toByteArray())));
	      QVariantList variantList;

	      /*
	      ** We don't need to store the URL's hash in
	      ** variantList.
	      */

	      variantList << url
			  << host
			  << bytes
			  << dateTime
			  << title
			  << visits
			  << description;
	      m_contentsList.append(variantList);

	      if(!dateStr.isEmpty())
		if(!dateTime.date().toString(Qt::ISODate).
		   contains(dateStr))
		  continue;

	      if(!searchStr.isEmpty())
		if(!(host.toLower().contains(searchStr) ||
		     title.toLower().contains(searchStr) ||
		     url.toString(QUrl::StripTrailingSlash).
		     toLower().contains(searchStr)))
		  continue;

	      QIcon icon;
	      QBuffer buffer;
	      QStandardItem *item = new QStandardItem(title);

	      item->setData(title, Qt::UserRole);
	      buffer.setBuffer(&bytes);
	      buffer.open(QIODevice::ReadOnly);

	      QDataStream in(&buffer);

	      in >> icon;
	      buffer.close();

	      if(icon.isNull())
		icon = dmisc::iconForUrl(url);

	      item->setIcon(icon);
	      item->setEditable(false);
	      m_model->setRowCount(m_model->rowCount() + 1);
	      m_model->setItem(m_model->rowCount() - 1, 0, item);
	      item = new QStandardItem(host);
	      item->setData(host, Qt::UserRole);
	      item->setEditable(false);
	      m_model->setItem(m_model->rowCount() - 1, 1, item);
	      item = new QStandardItem(url.toString(QUrl::StripTrailingSlash));
	      item->setData(url.toString(QUrl::StripTrailingSlash),
			    Qt::UserRole);
	      item->setEditable(false);
	      m_model->setItem(m_model->rowCount() - 1, 2, item);
	      item = new QStandardItem
		(dateTime.toString("MM/dd/yyyy hh:mm:ss AP"));

	      /*
	      ** Sorting dates is complicated enough. By
	      ** inserting dates in the below format,
	      ** we'll guarantee that the date sort will behave.
	      */

	      item->setData(dateTime.toString("yyyy/MM/dd hh:mm:ss"),
			    Qt::UserRole);
	      item->setEditable(false);
	      m_model->setItem(m_model->rowCount() - 1, 3, item);
	      item = new QStandardItem(visits);
	      item->setData(visits.toInt(), Qt::UserRole);
	      item->setEditable(false);
	      m_model->setItem(m_model->rowCount() - 1, 4, item);
	      item = new QStandardItem(urlHash);
	      m_model->setItem(m_model->rowCount() - 1, 5, item);
	      item = new QStandardItem(description);
	      m_model->setItem(m_model->rowCount() - 1, 6, item);
	    }

	slotSort(ui.history->horizontalHeader()->
		 sortIndicatorSection(),
		 ui.history->horizontalHeader()->
		 sortIndicatorOrder());
      }

    db.close();
  }

  QSqlDatabase::removeDatabase("history_window");
  statusBar()->showMessage(QString(tr("%1 item(s) - %2 selected")).
			   arg(m_model->rowCount()).
			   arg(ui.history->selectionModel()->
			       selectedRows(0).size()));
  ui.history->horizontalHeader()->setStretchLastSection(true);
  ui.history->setSelectionMode(QAbstractItemView::MultiSelection);

  while(!list.isEmpty())
    ui.history->selectRow(list.takeFirst().row());

  ui.history->setSelectionMode(QAbstractItemView::ExtendedSelection);
  ui.history->horizontalScrollBar()->setValue(hPos);
  ui.history->verticalScrollBar()->setValue(vPos);
  QApplication::restoreOverrideCursor();
}

void dhistory::slotItemSelectionChanged(void)
{
  if(isVisible())
    populate();
}

void dhistory::slotTextChanged(const QString &text)
{
  Q_UNUSED(text);
  populate();
}

void dhistory::closeEvent(QCloseEvent *event)
{
  m_timer->stop();
  saveState();
  QMainWindow::closeEvent(event);
}

void dhistory::saveState(void)
{
  if(isVisible())
    dooble::s_settings["historyWindow/geometry"] = saveGeometry();

  dooble::s_settings["historyWindow/splitterState"] =
    ui.splitter->saveState();
  dooble::s_settings["historyWindow/tableColumnsState5"] =
    ui.history->horizontalHeader()->saveState();

  QSettings settings;

  if(isVisible())
    settings.setValue("historyWindow/geometry", saveGeometry());

  settings.setValue("historyWindow/splitterState",
		    ui.splitter->saveState());
  settings.setValue("historyWindow/tableColumnsState5",
		    ui.history->horizontalHeader()->
		    saveState());
}

void dhistory::slotItemsSelected(const QItemSelection &selected,
				 const QItemSelection &deselected)
{
  Q_UNUSED(selected);
  Q_UNUSED(deselected);
  statusBar()->showMessage(QString(tr("%1 item(s) - %2 selected")).
			   arg(m_model->rowCount()).
			   arg(ui.history->selectionModel()->
			       selectedRows(0).size()));
}

void dhistory::keyPressEvent(QKeyEvent *event)
{
  if(event && event->key() == Qt::Key_Escape)
    close();
  else if(event && event->key() == Qt::Key_Delete)
    slotDeletePage();

  QMainWindow::keyPressEvent(event);
}

void dhistory::reencode(QProgressBar *progress)
{
  /*
  ** This method assumes that the respective container
  ** is as current as possible. It will use the container's
  ** contents to create new database entries using the newly-encoded
  ** data. Obsolete data will be deleted via the purge method.
  */

  if(progress)
    {
      progress->setMaximum(-1);
      progress->setVisible(true);
    }

  m_timer->stop();

  {
    QSqlDatabase db = QSqlDatabase::addDatabase("QSQLITE",
						"history_reencode");

    db.setDatabaseName(dooble::s_homePath +
		       QDir::separator() + "history.db");

    if(db.open())
      {
	QString hashType(dooble::s_settings.value("settingsWindow/hashType").
			 toString());
	QSqlQuery query(db);

	query.exec("PRAGMA synchronous = OFF");

	if(progress)
	  {
	    progress->setMaximum(m_contentsList.size());
	    progress->update();
	  }

	for(int i = 0; i < m_contentsList.size(); i++)
	  {
	    QVariantList list(m_contentsList.at(i));

	    query.prepare
	      ("INSERT OR REPLACE INTO history "
	       "(url, host, icon, last_visited, title, visits, "
	       "passphrase_hash, url_hash, description) "
	       "VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?)");
	    query.bindValue
	      (0,
	       dmisc::encodedString
	       (list.at(0).toUrl().toEncoded(QUrl::StripTrailingSlash),
		true).toBase64());
	    query.bindValue
	      (1,
	       dmisc::encodedString(list.at(1).toString().toUtf8(),
				    true).toBase64());
	    query.bindValue
	      (2,
	       dmisc::encodedString(list.at(2).toByteArray(),
				    true).toBase64());
	    query.bindValue
	      (3,
	       dmisc::encodedString(list.at(3).toDateTime().
				    toString(Qt::ISODate).
				    toUtf8(),
				    true).toBase64());

	    if(list.at(4).toString().trimmed().isEmpty())
	      query.bindValue
		(4,
		 dmisc::encodedString(list.at(0).toUrl().
				      toEncoded(QUrl::StripTrailingSlash),
				      true).toBase64());
	    else
	      query.bindValue
		(4,
		 dmisc::encodedString(list.at(4).toString().toUtf8(),
				      true).toBase64());

	    query.bindValue
	      (5,
	       dmisc::encodedString(list.at(5).toString().toAscii(),
				    true).toBase64());

	    QString hashType
	      (dooble::s_settings.value("settingsWindow/hashType").
	       toString());

	    query.bindValue
	      (6,
	       dmisc::passphraseHash(hashType));
	    query.bindValue
	      (7,
	       dmisc::hash(list.at(0).toUrl().
			   toEncoded(QUrl::StripTrailingSlash)).toHex().
	       constData());
	    query.bindValue
	      (8,
	       dmisc::encodedString(list.at(6).toString().toUtf8(),
				    true).toBase64());
	    query.exec();

	    if(progress)
	      progress->setValue(i + 1);
	  }
      }

    db.close();
  }

  QSqlDatabase::removeDatabase("history_reencode");
  purge();

  if(progress)
    progress->setVisible(false);

  m_timer->start();
}

void dhistory::slotBookmark(void)
{
  QModelIndexList list(ui.history->selectionModel()->selectedRows(0));

  while(!list.isEmpty())
    {
      QUrl url;
      QIcon icon;
      QString title("");
      QString description("");
      QStandardItem *item = m_model->item(list.first().row(), 0);

      if(item)
	{
	  icon = item->icon();
	  title = item->text();
	}

      item = m_model->item(list.first().row(), 2);

      if(item)
	url = QUrl::fromUserInput(item->text());

      item = m_model->item(list.first().row(), 6);

      if(item)
	description = item->text();

      if(icon.isNull())
	icon = dmisc::iconForUrl(url);

      emit bookmark(url, icon, title, description);
      list.takeFirst();
    }
}
